// RoffPhong.cpp: implementation of the RoffPhong class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RoffView.h"
#include "RoffPhong.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RoffPhong::RoffPhong()
{

    lightDirection = RoffVector(0, 0, -1);
    worldLight     = RoffVector(0, 0, -1);

	objOutsideRed   = 0.0;
	objOutsideGreen = 0.0;
	objOutsideBlue  = 1.0;

	objInsideRed   = 1.0;
	objInsideGreen = 0.0;
	objInsideBlue  = 0.0;

	ambientRed   = .75;
	ambientGreen = .75;
	ambientBlue  = .75;

	lightRed   = 1.0;
	lightGreen = 1.0;
	lightBlue  = 1.0;

	highlightRed   = .7;
	highlightGreen = .7;
	highlightBlue  = .7;

	ambientFactor  = .4;
	diffuseFactor  = .6;
	specularFactor = .7;

	shininess = 15.0;
}

RoffPhong::~RoffPhong()
{

}

/*
 * PRE: originalNormal.getLength() == 1
 *      AND
 *      viewDirection.getLength() == 1
 *      AND
 *      setWorldLight() has been called to set a usable value.
 */
COLORREF RoffPhong::Illuminate(const RoffVector& originalNormal, const RoffVector& viewDirection) const {

	RoffVector normal = originalNormal;
	double objDiffuseRed   = objOutsideRed;
	double objDiffuseGreen = objOutsideGreen;
	double objDiffuseBlue  = objOutsideBlue;

	// If we are viewing the backside of the object, use
	// the inside colors and the reversed normal.
	if (normal.Dot(viewDirection) < 0) {
		normal = normal * -1.0;
		objDiffuseRed   = objInsideRed;
		objDiffuseGreen = objInsideGreen;
		objDiffuseBlue  = objInsideBlue;
	}

	// N . L
	double diffFactor = normal.Dot(worldLight);
	// R = N * (N . L) - L
	RoffVector refVector = normal * (2 * diffFactor) - worldLight;
	// Normalize R
	refVector = refVector / refVector.getLength();
	// R . V
	double specFactor = refVector.Dot(viewDirection);

	// Begin with ambient color
	double redPart   = ambientRed * ambientFactor * objDiffuseRed; 
	double greenPart = ambientGreen * ambientFactor * objDiffuseGreen;
	double bluePart  = ambientBlue * ambientFactor * objDiffuseBlue;

	// Diffuse and specular portions only if the light is
	// on the object. (N . L is positive).
	if (diffFactor >= 0) {
		// Specular portion only if the reflection is coming
		// toward the viewer (R . V) is positive).
		if (specFactor >= 0) { 
			double specMultiplier = specularFactor * pow(specFactor, shininess);
			redPart   = redPart + specMultiplier * highlightRed * lightRed;
            greenPart = greenPart + specMultiplier * highlightGreen * lightGreen;
			bluePart  = bluePart + specMultiplier * highlightBlue * lightBlue;
        }
		double diffMultiplier = diffuseFactor * diffFactor;
        redPart   = redPart + diffMultiplier * objDiffuseRed * lightRed;
        greenPart = greenPart + diffMultiplier * objDiffuseGreen * lightGreen;
		bluePart  = bluePart + diffMultiplier * objDiffuseBlue * lightBlue;
	}
    
	// Clamp all three color portions.
	if (redPart > 1) {
		redPart = 1;
	}
	if (greenPart > 1) {
		greenPart = 1;
	}
	if (bluePart > 1) {
		bluePart = 1;
	}

	return GetColorSet(redPart, greenPart, bluePart);
}

void RoffPhong::setOutsideColor(const COLORREF newColor) {
	SetColorSet(newColor, objOutsideRed, objOutsideGreen, objOutsideBlue);
}

void RoffPhong::setInsideColor(const COLORREF newColor) {
	SetColorSet(newColor, objInsideRed, objInsideGreen, objInsideBlue);
}

void RoffPhong::setAmbientColor(const COLORREF newColor) {
	SetColorSet(newColor, ambientRed, ambientGreen, ambientBlue);
}

void RoffPhong::setDirectionalColor(const COLORREF newColor) {
	SetColorSet(newColor, lightRed, lightGreen, lightBlue);
}

void RoffPhong::setHighlightColor(const COLORREF newColor) {
	SetColorSet(newColor, highlightRed, highlightGreen, highlightBlue);
}

void RoffPhong::setLightDirection(const RoffVector& newDirection) {
    lightDirection = newDirection / newDirection.getLength();
}

void RoffPhong::setWorldLight(const RoffVector& newDirection) {
    worldLight = newDirection / newDirection.getLength();
}


COLORREF RoffPhong::getOutsideColor() const {
    return GetColorSet(objOutsideRed, objOutsideGreen, objOutsideBlue);
}

COLORREF RoffPhong::getInsideColor() const {
    return GetColorSet(objInsideRed, objInsideGreen, objInsideBlue);
}

COLORREF RoffPhong::getAmbientColor() const {
    return GetColorSet(ambientRed, ambientGreen, ambientBlue);
}

COLORREF RoffPhong::getDirectionalColor() const {
    return GetColorSet(lightRed, lightGreen, lightBlue);
}

COLORREF RoffPhong::getHighlightColor() const {
    return GetColorSet(highlightRed, highlightGreen, highlightBlue);
}

RoffVector RoffPhong::getLightDirection() const {
	return lightDirection;
}

RoffVector RoffPhong::getWorldLight() const {
	return worldLight;
}

void RoffPhong::SetColorSet(const COLORREF newColor, double& redPart, double& greenPart, double& bluePart) {

	int red = GetRValue(newColor);
	int green = GetGValue(newColor);
	int blue = GetBValue(newColor);

	redPart = (double)red / 255.0;
	greenPart = (double)green / 255.0;
	bluePart = (double)blue / 255.0;
}

COLORREF RoffPhong::GetColorSet(const double redPart, const double greenPart, const double bluePart) const {
	return RGB((int)(redPart * 255.0),
		       (int)(greenPart * 255.0),
			   (int)(bluePart * 255.0));
}