// RoffTriangleManager.cpp: implementation of the RoffTriangleManager class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "roffview.h"
#include "RoffTriangleManager.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RoffTriangleManager::RoffTriangleManager()
{

}

RoffTriangleManager::~RoffTriangleManager()
{
	for (int i = 0;i < numTriangles;i++) {
		delete triangles[i];
	}
}

void RoffTriangleManager::setPointArray(CArray<RoffVector, RoffVector&>* newPointArray) {
    pointArray = newPointArray;
}

void RoffTriangleManager::set2DPointArrays(CArray<double, double>* newPointXArray,
										   CArray<double, double>* newPointYArray) {
	pointXArray = newPointXArray;
	pointYArray = newPointYArray;
}

void RoffTriangleManager::tessellate(CArray<RoffPolygon3D, RoffPolygon3D&>* polyArray) {

	int i;
	int numPoints;
	RoffVector centroid;

	// Delete any old triangles that exist.
	for (i = 0;i < numTriangles;i++) {
		delete triangles[i];
	}
	numTriangles = 0;

	// Build the triangle list from the polygons in polyArray;
	int numPolys = polyArray->GetSize();
	for (i = 0;i < numPolys;i++) {
		//Find out how big the polygon is and set variables
		// accordingly.
		numPoints = (*polyArray)[i].getNumEdges();
		triangles.SetSize(triangles.GetSize() + numPoints - 2);
		// If we already have a triangle, just add it.
        if (numPoints == 3) {
            triangles[numTriangles] = new RoffTriangle((*polyArray)[i].getVertex(0),
				                                       (*polyArray)[i].getVertex(1),
											           (*polyArray)[i].getVertex(2),
											           pointArray);
			numTriangles++;
		} else {
		    // Find the centroid of the polgon.
		    centroid = (*pointArray)[(*polyArray)[i].getVertex(0)];
		    for (int j = 1;j < numPoints;j++) {
			    centroid = centroid + (*pointArray)[(*polyArray)[i].getVertex(j)];
			}
            centroid = centroid / numPoints;

			// Find the index of the point in the polygon closest to
			// the centroid.
			int indexClosest = 0;
			double distance;
			double minDistance = ((*pointArray)[(*polyArray)[i].getVertex(indexClosest)] - centroid).getSquaredLength();
			for (j = 1;j < numPoints;j++) {
                distance = ((*pointArray)[(*polyArray)[i].getVertex(j)] - centroid).getSquaredLength();
				if (distance < minDistance) {
					minDistance = distance;
					indexClosest = j;
				}
			}

			// Starting at that point, move around the polygon,
			// tearing off triangles.
			int secondPoint;
			int thirdPoint;
			for (j = 0; j < (numPoints - 2); j++) {
				secondPoint = indexClosest + j + 1;
				if (secondPoint >= numPoints) {
					secondPoint = secondPoint - numPoints;
				}
				thirdPoint = indexClosest + j + 2;
				if (thirdPoint >= numPoints) {
					thirdPoint = thirdPoint - numPoints;
				}
				triangles[numTriangles] = new RoffTriangle((*polyArray)[i].getVertex(indexClosest),
					                                       (*polyArray)[i].getVertex(secondPoint),
														   (*polyArray)[i].getVertex(thirdPoint),
														   pointArray);
				numTriangles++;
			}
		}
	}

	// Set an initial order that can be sorted once a From point
	// is picked.
	depthOrder.SetSize(numTriangles);
    for (i = 0;i < numTriangles;i++) {
		depthOrder[i] = i;
	}
}

void RoffTriangleManager::orderTriangles(RoffVector fromPoint) {

	int i;
	// Adjust all the triangles to have the correct depth for
	// the given from point.
    for (i = 0;i < numTriangles;i++) {
		triangles[i]->updateView(fromPoint);
	}

	// Sort depthOrder on the basis of triangles[index]->closer()
	// and triangles[index]->sameDepth().
	// Build a heap
	int heapsize = numTriangles;
    for (i = ((heapsize - 1) / 2); i >= 0; i --) {
        Heapify(i, heapsize);
	}

	// Sort from the heap.
	int tempIndex;
    for (i = numTriangles - 1; i > 0; i --) {
		tempIndex = depthOrder[0];
		depthOrder[0] = depthOrder[i];
		depthOrder[i] = tempIndex;
        heapsize --;
        Heapify(0, heapsize);
    }
}

/*
 * PRE:
 *     0 <= triangleIndex < getNumTriangles()
 */
RoffTriangle* RoffTriangleManager::get(int triangleIndex) {

	return triangles[depthOrder[triangleIndex]];
}

/*
 * PRE:
 *     0 <= triangleIndex < getNumTriangles()
 */
CPoint* RoffTriangleManager::getCPoints(int triangleIndex) {

	RoffTriangle* tri = triangles[depthOrder[triangleIndex]];

	CPoint* rval = new CPoint[3];

	rval[0].x = (int)((*pointXArray)[tri->getP1()] * screenFactor);
	rval[0].y = (int)(screenFactor - (*pointYArray)[tri->getP1()] * screenFactor);
	rval[1].x = (int)((*pointXArray)[tri->getP2()] * screenFactor);
	rval[1].y = (int)(screenFactor - (*pointYArray)[tri->getP2()] * screenFactor);
	rval[2].x = (int)((*pointXArray)[tri->getP3()] * screenFactor);
	rval[2].y = (int)(screenFactor - (*pointYArray)[tri->getP3()] * screenFactor);

	return rval;
}

void RoffTriangleManager::Heapify(int index, int heapsize) {

    int nodeBase = index * 2;
    int testOffset = 1;
    int largestIndex = index;

    while ( (nodeBase + testOffset < heapsize) && (testOffset <= 2) ){
        if (!(triangles[depthOrder[largestIndex]]->closer(*(triangles[depthOrder[nodeBase + testOffset]])))) {
            largestIndex = nodeBase + testOffset;
        }
        testOffset ++;
    }

    if ( largestIndex != index ) {
		int temp = depthOrder[index];
		depthOrder[index] = depthOrder[largestIndex];
		depthOrder[largestIndex] = temp;
        Heapify(largestIndex, heapsize);
    }
}

void RoffTriangleManager::rebuild() {
    for (int i = 0;i < numTriangles;i++) {
		triangles[i]->rebuild();
	}
}